/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

(function() {

/** Basic class for checking state of measurement */
class MeasureStateChecker {
    constructor(
        objAtts,        // Object containing values of object to create
        arrImportKeys   // Optional: Array of keys to import from `objAtts`
    ) {
        if (objAtts.value !== undefined)
            this.value = objAtts.value;
        if (objAtts.to !== undefined)
            this.to = objAtts.to;
        if (objAtts.inverted)
            this.inverted = true;
        if (arrImportKeys instanceof Array)
            arrImportKeys.forEach(key => {
                if (objAtts[key] !== undefined)
                    this[key] = objAtts[key];
            });
    }

    /**
    * Checks whether data triggers this rule
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        return true;
    }

    /**
    * Checks whether data triggers this rule with respect to `this.inverted`
    * @return boolean       Does data trigger this rule?
    */
    execute(
        data    // Object providing functions to read and write values
    ) {
        return this.exec(data) != this.inverted;
    }
}

/** Is check negated? */
MeasureStateChecker.prototype.inverted = false;

/** Most relevant value describing check */
MeasureStateChecker.prototype.value = undefined;

/** Eventually neccessary second value describing check */
MeasureStateChecker.prototype.to = undefined;

/** Types of MeasureStateChecker subclasses corresponding to type values */
MeasureStateChecker.types = {
    "accept": MeasureStateChecker
};

/**
* Create MeasureStateChecker instance described by object `data`
* @return MeasureStateChecker    Created instance (or null)
*/
MeasureStateChecker.fromObject = function(
    data    // Object describing wanted instance
) {
    if (data instanceof Object && MeasureStateChecker.types[data.type])
        return new MeasureStateChecker.types[data.type](data);
    return null;
}

/**
* Calculate value of a mathematical expression that can contain variable names
* @return Number    Result of calculation
*/
MeasureStateChecker.parseExpr = function(
    expression,     // String containing expression
    getValue        // Function to call to get value corresponding to variable name (including precision if second argument is `true`)
) {
    if (typeof expression !== 'string' && !(expression instanceof String))
        return { value: expression, precision: 0 };
    expression = expression.trim();
    if (!expression.length || expression == '0')
        return { value: 0, precision: 0 };
    switch (expression) {
        case 'true':
            return { value: true, precision: 1 };
        case 'false':
            return { value: false, precision: 1 };
        case 'Infinity':
            return { value: Infinity, precision: 1 };
    }
    if (expression.charAt(0) == '(' && expression.charAt(expression.length - 1) == ')')
        return MeasureStateChecker.parseExpr(expression.substring(1, expression.length - 1), getValue);
    let openbrackets = 0;
    let punkt = undefined;
    let strich = undefined;
    let pow = undefined;
    let cmp = undefined;
    for (let n = 0; n < expression.length; n++) {
        switch (expression.charAt(n)) {
            case '+':
            case '-':
                if (!openbrackets)
                    strich = n;
                break;
            case '*':
            case '/':
                if (!openbrackets)
                    punkt = n;
                break;
            case '^':
                if (!openbrackets)
                    pow = n;
                break;
            case '>':
                if (expression.charAt(n - 1) == '<')
                    break;
                // nobreak
            case '<':
                if (!openbrackets)
                    cmp = n;
                break;
            case '=':
                if (!openbrackets && expression.charAt(n + 1) == '=')
                    cmp = n;
                break;
            case '(':
                ++openbrackets;
                break;
            case ')':
                --openbrackets;
                break;
        }
    }
    const index = (cmp !== undefined) ? cmp : ((strich !== undefined) ? strich : ((punkt !== undefined) ? punkt : pow));
    if (index !== undefined) {
        const operator2chars = (index === cmp) && ['<>', '==', '<=', '=>'].indexOf(expression.substring(index, index + 2)) != -1;
        const operand1 = MeasureStateChecker.parseExpr(expression.substring(0, index), getValue);
        const operand2 = MeasureStateChecker.parseExpr(expression.substring(index + (operator2chars ? 2 : 1)), getValue);
        switch (expression.charAt(index)) {
            case '+':
                return { value: operand1.value + operand2.value, precision: operand1.precision + operand2.precision };
            case '-':
                return { value: operand1.value - operand2.value, precision: operand1.precision + operand2.precision };
            case '*':
                return { value: operand1.value * operand2.value, precision: Math.abs(operand1.value) * operand2.precision + Math.abs(operand2.value) * operand1.precision + operand1.precision * operand2.precision / 2 };
            case '/':
                return { value: operand1.value / operand2.value, precision: (Math.abs(operand1.value / operand2.value) * operand2.precision + operand1.precision) / Math.abs(Math.abs(operand2.value) - operand2.precision / 2) };
            case '^':
                const fktPow = (base, exponent) => Math.sign(base) * Math.pow(Math.abs(base), exponent);
                const vp = fktPow(operand1.value, operand2.value);
                const pps = [[operand1.value + operand1.precision, operand2.value + operand2.precision], [operand1.value - operand1.precision, operand2.value - operand2.precision]];
                pps.push([pps[0][0], pps[1][1]]);
                pps.push([pps[1][0], pps[0][1]]);
                const pp = pps.map(([base, exponent]) => Math.abs(fktPow(base, exponent) - vp)).filter(v => Number.isFinite(v));
                pp.sort((a, b) => a - b);
                return { value: vp, precision: pp.length ? pp[0] : 0 };
            case '<':
            case '>':
            case '=':
                const diffRaw = operand1.value - operand2.value;
                const precision = Math.max(operand1.precision, operand2.precision);
                const diff = precision ? (precision * Math.round(diffRaw / precision)) : diffRaw;
                switch (expression.substring(index, index + 2)) {
                    case '==':
                        return { value: diff == 0, precision: 1 };
                    case '<>':
                        return { value: diff != 0, precision: 1 };
                    case '<=':
                        return { value: diff <= 0, precision: 1 };
                    case '>=':
                        return { value: diff >= 0, precision: 1 };
                }
                return { value: (expression.charAt(index) == '<') ? (diff < 0) : (diff > 0), precision: 1 };
        }
    }
    if (expression.charAt(0) == '#')
        return { value: Math.round(-Math.log10(MeasureStateChecker.parseExpr(expression.substring(1), getValue).precision)), precision: 1 };
    if (expression.match(/^\d*.\d*$/))
        return { value: Number.parseFloat(expression), precision: Number.parseFloat(expression.replace(/\d/g, '0').replace(/0(?!.*0)/, '1')) };
    return getValue(expression, true);
}

/**
* Run checks described by `checkers`
* @return mixed     Return value of check
*/
MeasureStateChecker.run = function(
    checkers,       // Array of arrays of `MeasureStateChecker`'s JSON
    ssm,            // Responsible instance of `ScoStateMgr`
    simulation,     // Responsible instance of `MultimeterSimulation`
    values,         // Responsible instance of `RandomValues`
    unit            // Optional: Unit of input
) {
    let finalized = false;
    let result = undefined;
    let connected = null;
    let connectedExternal = null;
    let fine = 0;
    let msg = '';
    let leveldone = false;
    values.startNewCycle();
    const data = {
        getValue(name, includePrecision) { return values.read(name, includePrecision); },
        getMode() { return ssm.getValue('content.measureType'); },
        getConnected() { if (connected) return connected; return connected = simulation.createConnectionMatrix(); },
        getConnectedExternal() { if (connectedExternal) return connectedExternal; return connectedExternal = simulation.createConnectionMatrix(true); },
        getUnit() { return unit; },

        setDestroy(what) { ssm.setValue('content.destroyed', what); },
        setResult(value) { result = value; },
        setLeveldone(value) { leveldone = value; },
        setMessage(message) { msg = message; },
        setFine(v) { fine = v; },

        finalize() { finalized = true; }
    };

    for (let n = 0; n < checkers.length; ++n) {
        const checks = checkers[n];
        fine = 0;
        msg = '';
        result = undefined;
        if (!checks.reduce((prev, settings) => {
            if (prev)
                return true;
            const check = MeasureStateChecker.fromObject(settings);
            return (check && !check.execute(data)) ? true : false;
        }, false)) {
            if (fine)
                ssm.setValue('content.fine', ssm.getValue('content.fine') + fine);
            if (leveldone)
                ssm.setValue('content.levelsDone', 1 + ssm.getValue('content.levelsDone'));
            if (msg)
                ssm.setValue('content.message', msg);
            return result;
        }
    }
    return false;
}


class MeasureStateCheckerDestroy extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        data.setDestroy(this.value);
        data.finalize();
        return true;
    }
}


MeasureStateChecker.types['destroy'] = MeasureStateCheckerDestroy;


class MeasureStateCheckerReturn extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        data.setResult(MeasureStateChecker.parseExpr(this.value, data.getValue).value);
        data.finalize();
        return true;
    }
}


MeasureStateChecker.types['return'] = MeasureStateCheckerReturn;


class MeasureStateCheckerMode extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        return data.getMode() == MeasureStateChecker.parseExpr(this.value, data.getValue).value;
    }
}


MeasureStateChecker.types['mode'] = MeasureStateCheckerMode;


class MeasureStateCheckerConnected extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        const connected = data.getConnected()[this.value];
        if (this.to)
            return connected[this.to];
        return connected.reduce((prev, curr) => prev || curr, false);
    }
}


MeasureStateChecker.types['connected'] = MeasureStateCheckerConnected;


class MeasureStateCheckerConnectedExternal extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        const connected = data.getConnectedExternal();
        const connected2value = connected[this.value];
        if (!connected2value)
            return false;
        if (this.to)
            return connected2value[this.to];
        return Object.keys(connected2value).reduce((prev, key) => prev || connected2value[key], false);
    }
}


MeasureStateChecker.types['connectedExternal'] = MeasureStateCheckerConnectedExternal;


class MeasureStateCheckerMessage extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        data.setMessage((this.to || ScoTools.MSGTYPE_FAILED) + ':' + this.value);
        data.finalize();
        return true;
    }
}


MeasureStateChecker.types['message'] = MeasureStateCheckerMessage;


class MeasureStateCheckerEvaluate extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        const result = MeasureStateChecker.parseExpr(this.value, data.getValue);
        return result.value;
    }
}


MeasureStateChecker.types['evaluate'] = MeasureStateCheckerEvaluate;


class MeasureStateCheckerUnit extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        const raw = data.getUnit();
        const part = (!this.to && raw.length > 1 && 'pnµmcdkMG'.indexOf(raw.charAt(0)) != -1) ? raw.substring(1) : raw;
        return part == this.value;
    }
}


MeasureStateChecker.types['unit'] = MeasureStateCheckerUnit;


class MeasureStateCheckerFine extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        data.setFine(this.value);
        data.finalize();
        return true;
    }
}


MeasureStateChecker.types['fine'] = MeasureStateCheckerFine;


class MeasureStateCheckerLevelDone extends MeasureStateChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object providing functions to read and write values
    ) {
        data.setLeveldone(MeasureStateChecker.parseExpr(this.value, data.getValue).value);
        return true;
    }
}

MeasureStateChecker.types['leveldone'] = MeasureStateCheckerLevelDone;

window.MeasureStateChecker = MeasureStateChecker;
})();
