/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/***[ DigitalMultimeter.js ] *************************/
/** Class to enlive svg image of multimeter */
function DigitalMultimeter(
    svg,        // Element beeing or containing SVG to use
    digits,     // Array of digit information; each digit information is array of indizes of segments
    modes       // Optional: Array describing modes (see/defaults to DigitalMultimeter.stdModes)
) {
    this.modes = modes || DigitalMultimeter.stdModes;
    this.drehregler = svg.getElementsByClassName('drehregler')[0];
    const displaysegment = svg.getElementsByClassName('displaysegment');
    this._segmentFillInactive = displaysegment[0] ? displaysegment[0].style.fill : '';
    this.digits = digits.map(digit => digit.map(v => displaysegment[v]));
    const rotatestring = (this.drehregler && this.drehregler.hasAttribute('transform')) ? this.drehregler.getAttribute('transform') : '';
    const rotateparts = rotatestring.match(/rotate\((\d+\.?\d*)(.+)\)/);
    if (rotateparts) {
        this.showMode = function(mode) {
            while (mode < 0)
                mode += this.modes.length;
            if (this._mode == mode)
                return;
            const angle = (this.modecount - mode % this.modecount) * 360 / this.modecount;
            this.drehregler.setAttribute('transform', 'rotate(' + angle + rotateparts[2] + ')');
            this._mode = mode;
            if (typeof this.onchange == 'function')
                this.onchange.call(this, this.measureType, mode);
            this.simulate();
        }
    }
}

/** General modes of measurement */
DigitalMultimeter.MEASURE_OFF = 0;
DigitalMultimeter.MEASURE_VOLTAGE = 1;
DigitalMultimeter.MEASURE_CURRENT = 2;
DigitalMultimeter.MEASURE_RESISTANCE = 3;
DigitalMultimeter.MEASURE_BATTERY = 4;
DigitalMultimeter.MEASURE_INTERNAL = 5;

DigitalMultimeter.stdModes = [
    {
        desc: "OFF",
        predecimalcount: 3,
        type: DigitalMultimeter.MEASURE_OFF
    }, {
        desc: "1000V⎓",
        predecimalcount: 4,
        limit: 1000,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    }, {
        desc: "200V⎓",
        predecimalcount: 3,
        limit: 200,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    }, {
        desc: "20V⎓",
        predecimalcount: 2,
        limit: 20,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    }, {
        desc: "2000mV⎓",
        predecimalcount: 4,
        scale: 1000,
        limit: 2000,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    }, {
        desc: "200mV⎓",
        predecimalcount: 3,
        scale: 1000,
        limit: 200,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    }, {
        desc: "2000kΩ",
        predecimalcount: 4,
        scale: 0.001,
        limit: 2000,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "200kΩ",
        predecimalcount: 3,
        scale: 0.001,
        limit: 200,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "20kΩ",
        predecimalcount: 2,
        scale: 0.001,
        limit: 20,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "2000Ω",
        predecimalcount: 4,
        limit: 2000,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "200Ω",
        predecimalcount: 3,
        limit: 200,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "Ton- und Diodenzeichen",
        predecimalcount: 4,
        limit: 2000,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "TEMP °C",
        predecimalcount: 3,
        scale: 0.385,
        offset: -38.5,
        limit: 200,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "hFE",
        predecimalcount: 4,
        limit: 2000,
        type: DigitalMultimeter.MEASURE_RESISTANCE
    }, {
        desc: "10A⎓",
        predecimalcount: 2,
        limit: 10,
        type: DigitalMultimeter.MEASURE_CURRENT
    }, {
        desc: "200mA⎓",
        predecimalcount: 3,
        scale: 1000,
        limit: 200,
        type: DigitalMultimeter.MEASURE_CURRENT
    }, {
        desc: "20mA⎓",
        predecimalcount: 2,
        scale: 1000,
        limit: 20,
        type: DigitalMultimeter.MEASURE_CURRENT
    }, {
        desc: "2000µA⎓",
        predecimalcount: 4,
        scale: 1000000,
        limit: 2000,
        type: DigitalMultimeter.MEASURE_CURRENT
    }, {
        desc: "200V~",
        predecimalcount: 3,
        scale: 0.7,
        limit: 200,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    }, {
        desc: "750V~",
        predecimalcount: 4,
        scale: 0.7,
        limit: 750,
        type: DigitalMultimeter.MEASURE_VOLTAGE
    },
];
DigitalMultimeter.prototype.modes = DigitalMultimeter.stdModes;

/** Function to call whenever measurement has to be caluclated (return value is measurement in base units) */
DigitalMultimeter.prototype.onsimulate = function(measureType, mode) { return (measureType == DigitalMultimeter.MEASURE_OFF) ? undefined : 0; }

/** Function to call whenever mode changed */
DigitalMultimeter.prototype.onchange = function(measureType, mode) {};

/** Function to call whenever displayed value changes */
DigitalMultimeter.prototype.ondisplay = function(value, precision) {};

/**
* Start recalculation of measurement and display value
* @return undefined
*/
DigitalMultimeter.prototype.simulate = function() {
    const mode = this.modes[this._mode % this.modecount];
    const raw = this.onsimulate.call(this, mode.type, mode);
    const value = Number.isFinite(raw) ? (raw * (mode.scale || 1) + (mode.offset || 0)) : raw;
    if (Math.abs(value) >= (mode.limit || Infinity)) {
        this.showOverflow();
        return
    }
    if (!Number.isFinite(value)) {
        this.displayNothing();
        return;
    }
    if (typeof this.ondisplay == 'function') {
        const postdecimalcount = this.digits.length - 1 - (mode.predecimalcount || 0);
        const precision = Math.pow(10, -postdecimalcount) / (mode.scale || 1);
        this.ondisplay.call(this, raw, precision);
    }
    this.displayValue(value, mode.predecimalcount || 0);
}

/** Total number of available modes */
Object.defineProperty(DigitalMultimeter.prototype, 'modecount', { enumerable: true, get() { return this.modes.length; } });

/** Active mode */
DigitalMultimeter.prototype._mode = 0;
DigitalMultimeter.prototype.showMode = function(mode) {
    if (this._mode == mode)
        return;
    this._mode = mode;
    if (typeof this.onchange == 'function')
        this.onchange.call(this, mode);
    this.simulate();
}
Object.defineProperty(DigitalMultimeter.prototype, 'mode', {
    enumerable: true,
    get() { return this._mode; },
    set(v) { this.showMode(v); }
});

/** Current measure type */
Object.defineProperty(DigitalMultimeter.prototype, 'measureType', {
    enumerable: true,
    get() { return this.modes[this._mode].type; }
});

/** Shape of digits */
DigitalMultimeter.digitShapes = [0b0111111, 0b0000110, 0b1011011, 0b1001111, 0b1100110, 0b1101101, 0b1111101, 0b0000111, 0b1111111, 0b1101111];

/**
* Set shape to display by given digit
* @return undefined
*/
DigitalMultimeter.prototype.updateDigit = function(
    index,      // Index of digit
    value,      // Value to display (or not finte to display nothing)
    showDot     // Optional: Set true to display dot after this digit
) {
    const digit = this.digits[index];
    if (!(digit instanceof Array))
        return;
    digit.forEach((segment, index) => segment.style.fill = ((index < 7) ? (Number.isFinite(value) ? (DigitalMultimeter.digitShapes[value] & Math.pow(2, index)) : false) : showDot) ? '#000' : this._segmentFillInactive);
}

/**
* Indicate overflow
* @return undefined
*/
DigitalMultimeter.prototype.showOverflow = function() {
    this.digits.reduce((isFirst, digit, index) => {
        if (digit.length < 7 || !isFirst) {
            this.updateDigit(index, false, false);
            return isFirst;
        }
        this.updateDigit(index, 1, false);
        return false;
    }, true);
    if (typeof this.ondisplay == 'function')
        this.ondisplay.call(this, NaN, 0);
}

/**
* Display a value using all digits (or indicate overflow)
* @return undefined
*/
DigitalMultimeter.prototype.displayValue = function(
    value,              // Value to display
    predecimalcount     // Number of digits before decimal period
) {
    const postdecimalcount = this.digits.length - 1 - predecimalcount;
    const precision = Math.pow(10, -postdecimalcount);
    value = Math.round(value / precision) * precision;
    if (Math.abs(value) > 2 * Math.pow(10, predecimalcount - 1)) {
        this.showOverflow();
        return;
    }
    this.updateDigit(0, (value < 0) ? 8 : false, false);
    value = Math.abs(value) + precision / 10; // Should fix unprecision issues of javascript's floating point arithmetics combined with Math.floor
    let stufe = 1;
    for (let n = 0; n < predecimalcount; ++n) {
        this.updateDigit(predecimalcount - n, (n ? (value < stufe) : false) ? false : Math.floor(value / stufe) % 10, !n);
        stufe *= 10;
    }
    stufe = 1;
    for (let n = 1; n + predecimalcount < this.digits.length; ++n) {
        stufe *= 10;
        this.updateDigit(predecimalcount + n, Math.floor(value * stufe) % 10, false);
    }
}

/**
* Clear display
* @return undefined
*/
DigitalMultimeter.prototype.displayNothing = function() {
    this.digits.forEach((digit, index) => this.updateDigit(index, false, false));
    if (typeof this.ondisplay == 'function')
        this.ondisplay.call(this, NaN, 0);
}

/***[ Connections.js ] *************************/
/** Class to manage and render connections */
function Connections(
    svg,        // SVG element to use
    internal    // Optional: Array of pairs of internally connected indizes (e.g. [ [0, 1], [3, 5] ])
) {
    this.node = svg;
    this._connections = [];
    this._internal = [];
    internal
        .map(arr => arr.filter(v => Number.isFinite(v)))
        .forEach(arr => arr.forEach((n, index) => index && this._internal.push([arr[0], n])));
}

/** Information on external connections */
Connections.prototype._connections = [];

/** Number of direct external connections */
Object.defineProperty(Connections.prototype, 'count', {
    enumerable: true,
    get() { return this._connections.length; }
});

/** Internally connected (e.g. [ [0, 1], [3, 5] ] if 0 and 1 are connected and also 3 and 5) */
Connections.prototype._internal = [];

/** Colors to use for connections */
Connections.prototype.colors = ['#edd400', '#73d216', '#75507b', '#c17d11', '#d3d7cf', '#555753', '#f57900', '#c00', '#3465a4'];
/** Index of next color to use */
Connections.prototype.nextColor = 0;

/** Function to call whenever direct external connections change */
Connections.prototype.onchange = function() {}

/**
* Check if there is a direct external connection of the connectors with given indizes
* @return boolean       Are theese connectors directly connected? (A connector is always connected to itself)
*/
Connections.prototype.isConnected = function(
    index1,     // Index of one connector in question (order is irrelevant)
    index2      // Index of one connector in question (order is irrelevant)
) {
    if (index1 > index2)
        [index1, index2] = [index2, index1];
    return this._connections.reduce((prev, c) => prev || (c.index1 == index1 && c.index2 == index2), index1 == index2);
}

/**
* Add a direct external connection unless connectors given by there indizes are already connected
* @return undefined
*/
Connections.prototype.add = function(
    x1,         // Horizontal coordinate of one connector (in local svg units)
    y1,         // Vertical coordinate of one connector (in local svg units)
    index1,     // Index of one connector in question (order is irrelevant)
    x2,         // Horizontal coordinate of one connector (in local svg units)
    y2,         // Vertical coordinate of one connector (in local svg units)
    index2,     // Index of one connector in question (order is irrelevant)
    color       // Optional: Color (string!) to use (or false to use default color)
) {
    if (index1 > index2)
        [x1, y1, index1, x2, y2, index2] = [x2, y2, index2, x1, y1, index1];
    if (this.isConnected(index1, index2))
        return;
    const nodeBg = this.node.ownerDocument.createElementNS('http://www.w3.org/2000/svg', 'line');
    nodeBg.setAttribute('x1', x1);
    nodeBg.setAttribute('y1', y1);
    nodeBg.setAttribute('x2', x2);
    nodeBg.setAttribute('y2', y2);
    const nodeFg = nodeBg.cloneNode(false);
    nodeBg.setAttribute('class', 'kabelback');
    nodeFg.setAttribute('class', 'kabelfront');
    let useColor = false;
    if (typeof color == 'string' || color instanceof String)
        useColor = color;
    if (color === undefined) {
        if (this.colors[this.nextColor])
            useColor = this.colors[this.nextColor];
        this.nextColor = (this.nextColor + 1) % this.colors.length;
    }
    if (useColor)
        nodeFg.setAttribute('style', 'stroke:' + useColor);
    this.node.appendChild(nodeBg);
    this.node.appendChild(nodeFg);
    this._connections.push({ index1, index2, nodeBg, nodeFg, color: useColor });
    if (typeof this.onchange == 'function')
        this.onchange.call(this);
}

/**
* Remove direct external connection of two connectors given by there indizes (if there is any connection)
* @return undefined
*/
Connections.prototype.remove = function(
    index1,     // Index of one connector in question (order is irrelevant)
    index2      // Index of one connector in question (order is irrelevant)
) {
    if (index1 > index2)
        [index1, index2] = [index2, index1];
    let change = false;
    this._connections = this._connections.filter(c => {
        if (c.index1 != index1 || c.index2 != index2)
            return true;
        this.node.removeChild(c.nodeFg);
        this.node.removeChild(c.nodeBg);
        change = true;
        return false;
    });
    if (change && typeof this.onchange == 'function')
        this.onchange.call(this);
}

/**
* Remove all direct external connections
* @return undefined
*/
Connections.prototype.clear = function() {
    if (!this._connections.length)
        return;
    this._connections.forEach(c => {
        this.node.removeChild(c.nodeFg);
        this.node.removeChild(c.nodeBg);
    });
    this._connections = [];
    if (typeof this.onchange == 'function')
        this.onchange.call(this);
}

/**
* Create array of indizes of all connectors beeing connected to one connector
* @return Array     Array of indizes
*/
Connections.prototype.getConnectedTo = function(
    index,              // Index of connector in question
    excludeInternal,    // Optional: Set to true if internal connections are not to be taken into account
    onlyDirect          // Optional: Set to true if indirect connections are not to be taken into account
) {
    const result = [index];
    let unused = this._connections.slice();
    if (!excludeInternal)
        unused = unused.concat(this._internal.map(([index1, index2]) => ({ index1, index2 })));
    let lastLength = 0;
    while (lastLength != unused.length) {
        lastLength = unused.length;
        unused = unused.filter(c => {
            if (result.indexOf(c.index1) != -1) {
                if (result.indexOf(c.index2) == -1)
                    result.push(c.index2);
                return false;
            }
            if (result.indexOf(c.index2) != -1) {
                result.push(c.index1);
                return false;
            }
            return true;
        });
        if (onlyDirect)
            break;
    }
    return result;
}

/**
* Ensure there are only direct external connections described by `connections`
* @return undefined
*/
Connections.prototype.setConnections = function(
    connections     // Array of objects with attributes like this.add parameters describing all wanted connections
) {
    let changed = false;
    const realOnchange = this.onchange;
    if (typeof this.onchange == 'function')
        this.onchange = () => (changed = true);
    const allowed = connections.map(({ index1, index2 }) => Math.min(index1, index2) + '-' + Math.max(index1, index2));
    this._connections.forEach(({ index1, index2 }) => {
        if (allowed.indexOf(Math.min(index1, index2) + '-' + Math.max(index1, index2)) == -1)
            this.remove(index1, index2);
    });
    connections.forEach(({ x1, y1, index1, x2, y2, index2, color }) => this.add(x1, y1, index1, x2, y2, index2, color));
    if (typeof realOnchange == 'function') {
        this.onchange = realOnchange;
        if (changed)
            this.onchange.call(this);
    }
}

/**
* Create string containing the most relevant information on direct external connections
* @return String    Data as comma seperated JSON strings of three values per connection (index1, index2 and color)
*/
Connections.prototype.toString = function() {
    const data = [];
    this._connections.forEach(({ index1, index2, color }) => {
        data.push(index1);
        data.push(index2);
        data.push(color);
    });
    const raw = JSON.stringify(data);
    return raw.substring(1, raw.length - 1);
}

/***[ MultimeterSimulation.js ] *************************/
function MultimeterSimulation(
    svgNode,            // SVG element containing multimeter, layer with button places etc.
    digitIndizes,       // Array of arrays containing index of corresponding segment
    connectorNames,     // Array of internal names of all connectors
    internalConnected,  // Optional: Array of pairs of internally connected connectors' names (elements of `connectorNames`!)
    modes               // Optional: Array describing modes of multimeter (see/defaults to DigitalMultimeter.stdModes)
) {
    this.node = svgNode;
    this.multimeter = new DigitalMultimeter(svgNode, digitIndizes, modes);
    const btns = new SvgPartsButtons(
        svgNode.getElementsByClassName('btnplaces')[0],
        svgNode.parentNode,
        title => title.startsWith(this.labelStartMode) ? 'Diesen Messbereich auswählen' : 'Hier anschließen'
    );
    this.connectorNames = (connectorNames instanceof Array) ? connectorNames.slice() : [];
    this.connections = new Connections(svgNode, internalConnected.map(pair => pair.map(name => this.getConnectorIndex(name))));
    this.connectorPositions = this.connectorNames.map(name => MultimeterSimulation.getRectCenter(btns.getPartByTitle(name)));

    this.multimeter.onsimulate = (type, mode) => {
        if (this.gameover)
            return false;
        const result = this.onsimulate.call(this, type, mode);
        return this.gameover ? false : result;
    }
    this.multimeter.onchange = (measureType, mode) => {
        if (typeof this.onmultimetermodechange == 'function')
            this.onmultimetermodechange.call(this, measureType, mode);
    };
    this.multimeter.ondisplay = (value, precision) => {
        if (typeof this.onmultimeterdisplay == 'function')
            this.onmultimeterdisplay.call(this, value, precision);
    };
    this.connections.onchange = () => {
        if (typeof this.onconnectionchange == 'function')
            this.onconnectionchange.call(this, this.connections.toString());
    }

    svgNode.addEventListener('mousemove', ev => {
        if (!this._preview)
            return;
        const [x, y] = this.client2local(ev.clientX, ev.clientY);
        this._updateConnectionPreview(x, y);
    });

    btns.onenter = (label, rect) => {
        if (!this._preview)
            return;
        if (label.startsWith(this.labelStartMode)) {
            this._updateConnectionPreview();
            return;
        }
        const [x, y] = MultimeterSimulation.getRectCenter(rect);
        this._updateConnectionPreview(x, y, label);
    }

    btns.onclick = (label, rect) => {
        if (this.gameover) {
            alert('Sinnlos! Da du bereits ' + this.gameover + ' zerstört hast, kannst du die Messung nicht mehr durchführen…');
            return;
        }
        if (!label.startsWith(this.labelStartMode)) {
            const [x, y] = MultimeterSimulation.getRectCenter(rect);
            this._buildConnection(x, y, label);
            return;
        }
        this.stopConnectionCreation();
        this.setMeasureMode(label.substring(this.labelStartMode.length));
    }
}

/** Function to call whenever measurement has to be caluclated (return value is measurement in base units) */
MultimeterSimulation.prototype.onsimulate = function(measureType, mode) { return (measureType == DigitalMultimeter.MEASURE_OFF) ? undefined : 0; }

/** Function to call whenever multimeter's measure mode changed */
DigitalMultimeter.prototype.onmultimetermodechange = function(measureType, mode) {};

/** Function to call whenever displayed value changes */
DigitalMultimeter.prototype.onmultimeterdisplay = function(value, precision) {};

/** Function to call whenever direct external connections changed */
DigitalMultimeter.prototype.onconnectionchange = function(serializedConnections) {};

/** First charecters of every string indicating a measure mode */
MultimeterSimulation.prototype.labelStartMode = 'Messbereich ';

/** Is simulation over because user has crashed something? If yes `gameover`'s value describes crashed part */
MultimeterSimulation.prototype._gameover = false;
Object.defineProperty(MultimeterSimulation.prototype, 'gameover', {
    enumerable: true,
    get() { return this._gameover; },
    set(v) {
        if (!v)
            this._gameover = false;
        if (!v || this._gameover)
            return;
        this._gameover = v;
        this.stopConnectionCreation();
        this.multimeter.displayNothing();
    }
});

/** Used instance of DigitalMultimeter */
MultimeterSimulation.prototype.multimeter = null;

/** Used instance of Connections */
MultimeterSimulation.prototype.connections = null;

/** Array of internal names of all connectors */
MultimeterSimulation.prototype.connectorNames = [];

/** Positions of corresponding connectors (items are arrays containing x and y value) */
MultimeterSimulation.prototype.connectorPositions = [];

/**
* Start recalculation of measurement and display value
* @return undefined
*/
MultimeterSimulation.prototype.simulate = function() {
    this.multimeter.simulate();
}

/**
* Calculate horizontal and vertical position of a rect's center in local svg units
* @return Array     [x, y]
*/
MultimeterSimulation.getRectCenter = function(
    rect    // SVGRectElement in question
) {
    if (!rect)
        return [0, 0];
    const x = parseFloat(rect.getAttribute('x')) + 0.5 * parseFloat(rect.getAttribute('width'));
    const y = parseFloat(rect.getAttribute('y')) + 0.5 * parseFloat(rect.getAttribute('height'));
    return [x, y];
}

/**
* Convert coordinate relative to viewport to coordinate in local svg units relative to svg image
* @return Array     [x, y]
*/
MultimeterSimulation.prototype.client2local = function(
    clientX,
    clientY
) {
    const svgRect = this.node.getBoundingClientRect();
    const realHeight = parseFloat(this.node.getAttribute('height'));
    const realWidth = parseFloat(this.node.getAttribute('width'));
    const offset = (svgRect.width - realWidth / realHeight * svgRect.height) / 2;
    const x = Math.min(Math.max(0, clientX - svgRect.left - offset) / (svgRect.width - 2 * offset), 1) * realWidth;
    const y = (clientY - svgRect.top) * realHeight / svgRect.height;
    return [x, y];
}

/**
* Get number used to describe connector by this.connections
* @return Number    Index of connector
*/
MultimeterSimulation.prototype.getConnectorIndex = function(
    name        // Internal name of connector
) {
    return this.connectorNames.indexOf(name);
}

/**
* Check if there is a direct external connection of the connectors with given names
* @return boolean       Are theese connectors directly connected? (A connector is always connected to itself)
*/
MultimeterSimulation.prototype.isConnected = function(
    name1,      // Name of one connector in question (order is irrelevant)
    name2       // Name of one connector in question (order is irrelevant)
) {
    return this.connections.isConnected(this.getConnectorIndex(name1), this.getConnectorIndex(name2));
}

/**
* Create array of names of all connectors beeing connected to one connector
* @return Array     Array of indizes
*/
MultimeterSimulation.prototype.getConnectedTo = function(
    name,               // Name of connector in question
    excludeInternal,    // Optional: Set to true if internal connections are not to be taken into account
    onlyDirect          // Optional: Set to true if indirect connections are not to be taken into account
) {
    return this.connections.getConnectedTo(this.getConnectorIndex(name), excludeInternal, onlyDirect).map(index => this.connectorNames[index]);
}

/**
* Create an object in a way that `this.createConnectionMatrix()[name1][name2]` is `true` if connectors with internal names `name1` and `name2` are connected and `false` otherwise
* @return Object    Object describing connections
*/
MultimeterSimulation.prototype.createConnectionMatrix = function(
    excludeInternal,    // Optional: Set to true if internal connections are not to be taken into account
    onlyDirect          // Optional: Set to true if indirect connections are not to be taken into account
) {
    const result = {};
    this.connectorNames.forEach((name1, index1) => {
        result[name1] = {};
        const linkedTo = this.connections.getConnectedTo(index1, excludeInternal, onlyDirect);
        this.connectorNames.forEach((name2, index2) => result[name1][name2] = (linkedTo.indexOf(index2) != -1));
    });
    return result;
}

/** Data used to preview connection */
MultimeterSimulation.prototype._preview = null;

/**
* Abort creation of a connection (if any)
* @return undefined
*/
MultimeterSimulation.prototype.stopConnectionCreation = function() {
    if (!this._preview)
        return;
    if (this._preview.node && this._preview.node.parentNode)
        this._preview.node.parentNode.removeChild(this._preview.node);
    this._preview = null;
}

/**
* Start creation of a new connection and enable its preview
* @return undefined
*/
MultimeterSimulation.prototype._startConnectionPreview = function(
    x,      // Horizontal position of connection's start in local svg units
    y,      // Vertical position of connection's start in local svg units
    name    // Internal name of connector
) {
    this.stopConnectionCreation();
    const nodePreview = this.node.ownerDocument.createElementNS('http://www.w3.org/2000/svg', 'line');
    this._preview = { node: nodePreview, x, y, name, index: this.getConnectorIndex(name) };
    nodePreview.setAttribute('x1', x);
    nodePreview.setAttribute('y1', y);
    nodePreview.setAttribute('x2', x);
    nodePreview.setAttribute('y2', y);
    nodePreview.setAttribute('class', 'preview');
    this.node.appendChild(nodePreview);
}

/**
* Update connection preview
* @return undefined
*/
MultimeterSimulation.prototype._updateConnectionPreview = function(
    x,          // Optional: Current horizontal position of connection's end in local svg units
    y,          // Optional: Current vertical position of connection's end in local svg units
    name        // Optional: Internal name of current connector
) {
    if (!this._preview)
        return;
    const nodePreview = this._preview.node;
    nodePreview.setAttribute('x2', Number.isFinite(x) ? x : this._preview.x);
    nodePreview.setAttribute('y2', Number.isFinite(y) ? y : this._preview.y);
    if (!name || !(Number.isFinite(x) || Number.isFinite(y))) {
        nodePreview.setAttribute('class', 'preview');
        return;
    }
    const index = this.getConnectorIndex(name);
    nodePreview.setAttribute('class', (index == this._preview.index) ? 'preview' : (this.connections.isConnected(this._preview.index, index) ? 'preview-del' : 'preview-add'));
}

/**
* Create a connection (if start is already given) or set its start
* @return undefined
*/
MultimeterSimulation.prototype._buildConnection = function(
    x,      // Horizontal position of connection's connector in local svg units
    y,      // Vertical position of connection's connector in local svg units
    name    // Internal name of connector
) {
    if (!this._preview)
        return this._startConnectionPreview(x, y, name);
    if (this._preview.name != name)
        this._addConnection(this._preview.x, this._preview.y, this._preview.index, x, y, this.getConnectorIndex(name));
    this.stopConnectionCreation();
}

/**
* Create and display an external connection or delete it (if it is already there)
* @return undefined
*/
MultimeterSimulation.prototype._addConnection = function(
    x1,         // Horizontal coordinate of one connector (in local svg units)
    y1,         // Vertical coordinate of one connector (in local svg units)
    index1,     // Index of one connector in question (order is irrelevant)
    x2,         // Horizontal coordinate of one connector (in local svg units)
    y2,         // Vertical coordinate of one connector (in local svg units)
    index2      // Index of one connector in question (order is irrelevant)
) {
    if (this.connections.isConnected(index1, index2))
        this.connections.remove(index1, index2);
    else
        this.connections.add(x1, y1, index1, x2, y2, index2);
    this.multimeter.simulate();
    return;
}

/** Timeout of step by step mode switching */
MultimeterSimulation.prototype._timeoutChangeMode = false;

/**
* Change measure mode
* @return undefined
*/
MultimeterSimulation.prototype.setMeasureMode = function(
    label,          // Label of new mode (or mode number)
    directly        // Optional: Set to true if mode should change instantly
) {
    const modeNew = Number.isFinite(label) ? label : this.multimeter.modes.map(({ desc }) => desc).indexOf(label);
    if (this._timeoutChangeMode) {
        clearTimeout(this._timeoutChangeMode);
        this._timeoutChangeMode = false;
    }
    if (directly && modeNew != this.multimeter.mode) {
        this.multimeter.mode = modeNew;
        return;
    }
    if (modeNew != -1 && modeNew != this.multimeter.mode) {
        const diffInc = (this.multimeter.modecount + modeNew - this.multimeter.mode) % this.multimeter.modecount;
        const diffDec = (this.multimeter.modecount + this.multimeter.mode - modeNew) % this.multimeter.modecount;

        const changeMode = (modeNew, step) => {
            this._timeoutChangeMode = false;
            if (this.gameover)
                return;
            this.multimeter.mode = (this.multimeter.modecount + this.multimeter.mode + step) % this.multimeter.modecount;
            this._timeoutChangeMode = (this.multimeter.mode == modeNew) ? false : setTimeout(() => changeMode(modeNew, step), 250);
        }
        changeMode(modeNew, (diffDec < diffInc) ? -1 : +1);
    }
}

/**
* Ensure there are only direct external connections described by `arr`
* @return undefined
*/
MultimeterSimulation.prototype.setConnections = function(
    arr         // Array containing per connection two indizes and a color
) {
    const data = [];
    while (arr.length >= 2) {
        const index1 = arr.shift();
        const index2 = arr.shift();
        const color = arr.shift();
        const pos1 = this.connectorPositions[index1];
        const pos2 = this.connectorPositions[index2]
        if (!(pos1 instanceof Array && pos2 instanceof Array))
            continue;
        data.push({ x1: pos1[0], y1: pos1[1], index1, x2: pos2[0], y2: pos2[1], index2, color });
    }
    this.connections.setConnections(data);
}

/**
* Get number of direct external connections
* @return Number    Number of direct external connections
*/
MultimeterSimulation.prototype.getConnectionCount = function() {
    return this.connections.count;
}

/***[ RandomValue.js ] *************************/
/** Manages and eventually recalulates a random value */
function RandomValue(
    settings,       // Object describing settings of this random value (see window.TASK.randomValues items)
    ssm             // Instance of ScoStateMgr; current value is stored as `"content.value." + settings.name`
) {
    this.settings = settings;
    this.ssm = ssm;
    this.precision = settings.step || 1;
    ssm.addListener('content.value.' + this.settings.name, () => this.isValid() || this.recalculate(), true);
}

/** Object describing settings of this random value (see window.TASK.randomValues items) */
RandomValue.prototype.settings = { 'name': '' };

/** Instance of ScoStateMgr; current value is stored as `"content.value." + this.settings.name` */
RandomValue.prototype.ssm = null;

/** How accurate the stored value is */
RandomValue.prototype.precision = 1;

/** Current value */
Object.defineProperty(RandomValue.prototype, 'current', {
    enumerable: true,
    get() { return this.ssm.getValue('content.value.' + this.settings.name); },
    set(v) { this.ssm.setValue('content.value.' + this.settings.name, v); }
});

/** Should value be recalculated before use? */
RandomValue.prototype._needRecalc = false;

/**
* Start a new useage cycle and eventually mark this value to be recalculated before use
* @return undefined
*/
RandomValue.prototype.startNewCycle = function() {
    this._needRecalc = this.settings.floating || false;
}

/**
* Check if this value has given name
* @return boolean   Does name fit?
*/
RandomValue.prototype.hasName = function(
    name    // Wanted name
) {
    return (this.settings.name == name);
}

/**
* Check if this value has given import type
* @return boolean   Does import type fit?
*/
RandomValue.prototype.isImport = function(
    type    // Import type's name
) {
    return (this.settings['import'] || '') == type;
}

/**
* Recalculate value (if needed) and return it
* @return Number    Current value
*/
RandomValue.prototype.read = function() {
    if (this._needRecalc) {
        this.recalculate();
        this._needRecalc = false;
    }
    return this.current;
}

/**
* Check if current value fits settings
* @return boolean       Is current value okay?
*/
RandomValue.prototype.isValid = function() {
    const step = (this.settings.step || 1);
    const current = Math.round(this.current / step) * step;
    if (this.settings.exclude instanceof Array && this.settings.exclude.reduce((prev, curr) => prev || Math.abs(curr - current) < step, false))
        return false;
    return Math.abs(current - (this.settings.mean || 0)) <= (this.settings.maxdiff || Infinity);
}

/**
* Force recalulation of this value (and store new value)
* @return undefined
*/
RandomValue.prototype.recalculate = function() {
    const mean = (this.settings.mean || 0);
    const maxdiff = Number.isFinite(this.settings.maxdiff) ? this.settings.maxdiff : Number.MAX_SAFE_INTEGER;
    const step = (this.settings.step || 1);
    const exclude = (this.settings.exclude instanceof Array) ? this.settings.exclude.filter(v => Math.abs(v - mean) <= maxdiff) : [];
    exclude.sort((a, b) => a - b);
    const value = this.settings.linear ? RandomValue._recalculateLinear(mean, maxdiff, step, exclude) : RandomValue._recalculateNonlinear(mean, maxdiff, step, exclude);
    this.precision = step;
    this.current = value;
}

/**
* Force recalculation of this value using linear distribution
* @return Number        New value
*/
RandomValue._recalculateLinear = function(
    mean,       // Mean value of distribution
    maxdiff,    // Maximum allowed absolute deviation of result from `mean`
    step,       // Minimal allowed absolute deviation of two possible results
    exclude     // Ascending sorted array of otherwise possible results to exclude
) {
    const rangelength = Math.floor(2 * maxdiff / step) + 1 - exclude.length;
    const value = mean - maxdiff + Math.floor(Math.random() * rangelength) * step;
    return exclude.reduce((value, exclude) => (value >= exclude) ? (value + step) : value, value);
}

/**
* Force recalculation of this value using a non linear distribution
* @return Number        New value
*/
RandomValue._recalculateNonlinear = function(
    mean,       // Mean value of distribution
    maxdiff,    // Maximum allowed absolute deviation of result from `mean`
    step,       // Minimal allowed absolute deviation of two possible results
    exclude     // Ascending sorted array of otherwise possible results to exclude
) {
    const exponent = 1.6;
    const rangelength = Math.floor(2 * maxdiff / step) + 1;
    const relativehalfstep = 1 / rangelength;
    const excludeExt = exclude.map(notAllowed => {
        const centered = 2 * Math.floor((notAllowed - mean) / step + 0.5) / rangelength;
        const start = centered - relativehalfstep;
        const end = centered + relativehalfstep;
        const pos = Math.sign(start) * Math.pow(Math.abs(start), 1 / exponent);
        const length = Math.abs(Math.pow(Math.abs(end), 1 / exponent) - Math.sign(end) * pos);
        return { pos, length };
    });
    const random = Math.random() * (2 - excludeExt.reduce((prev, { length }) => prev + length, 0)) - 1;
    const pos = excludeExt.reduce((pos, exclude) => (pos >= exclude.pos) ? ( pos + exclude.length) : pos, random);
    const stepsdiff = Math.floor(Math.pow(Math.abs(pos), exponent) * rangelength / 2 + 0.5)
    const result = mean + Math.sign(pos) * stepsdiff * step;
    return result;
}

/***[ RandomValues.js ] *************************/
/** Manages set of RandomValue instances */
function RandomValues(
    data,   // Array of settings of all RandomValue instances (see `window.TASK.randomValues`)
    ssm     // Instance of ScoStateMgr; current value is stored as `"content.value." + settings.name`
) {
    this.values = data.map(settings => new RandomValue(settings, ssm));
}

/** All RandomValue instances */
RandomValues.prototype.values = [];

/**
* Tell all floating values that they need to be recalculated
* @return undefined
*/
RandomValues.prototype.startNewCycle = function() {
    this.values.forEach(obj => obj.startNewCycle());
}

/**
* Read value with given name; recalculate it before if needed
* @return mixed     Read value (if `includePrecision == false`) or object like { value: 'read value', precision: 'max diff from real value' }
*/
RandomValues.prototype.read = function(
    name,               // Name of value in question
    includePrecision    // Optional: Also include precision in return value?
) {
    for (let n = 0; n < this.values.length; ++n) {
        if (this.values[n].hasName(name))
            return includePrecision ? { value: this.values[n].read(), precision: this.values[n].precision } : this.values[n].read();
    }
    return 0;
}

/**
* Set all values with given import type to a value
* @return undefined
*/
RandomValues.prototype.importValue = function(
    type,       // Import type
    value,      // New value
    precision   // `value`'s precision
) {
    this.values.forEach(item => {
        if (!item.isImport(type))
            return;
        item.precision = precision;
        item.current = value;
    });
}

/***[ MeasureResults.js ] *************************/
function MeasureResults(
    parentNode,
    levels,
    ssm,
    simulation,
    values
) {
    this.simulation = simulation;
    this.values = values;
    this.ssm = ssm;
    this.node = createHtml('form', parentNode, { 'method': 'post', 'action': location.href, "class": "measurements" });
    this.node.addEventListener('submit', ev => {
        ev.preventDefault();
        this.check();
    });
    const fieldset = createHtml('fieldset', this.node);
    createHtml('legend', fieldset, null, 'Messungen');
    this.msgPlace = new ScoTools.MessagePlace();
    this.msgPlace.appendAsChild(fieldset);
    this.listnode = createHtml('ol', fieldset);
    this.items = ((levels instanceof Array) ? levels : []).map(data => new MeasureResults.Item(this, data));
    ssm.addListener('content.levelsDone', levelsDone => {
        ssm.setValue('content.message', '');
        const failed = ssm.getValue('content.destroyed') ? true : false;
        this.items.forEach((item, index) => item.updateState((index < levelsDone) ? 1 : ((index == levelsDone) ? (failed ? 2 : 0) : -1)));
        this.check(true);
    }, true);
    ssm.addListener('content.destroyed', destroyed => {
        const failed = destroyed ? true : false;
        const levelsDone = ssm.getValue('content.levelsDone');
        this.items.forEach((item, index) => item.updateState((index < levelsDone) ? 1 : ((index == levelsDone) ? (failed ? 2 : 0) : -1)));
    }, false);
    ssm.addListener('content.inputs', s => {
        try {
            data = JSON.parse('[' + s + ']');
            data.forEach((value, index) => this.items[index] && this.items[index].setValue(value));
            this.check(true);
        } catch(err) {}
    });
    ssm.addListener('content.message', msg => {
        if (!msg) {
            this.msgPlace.clear();
            return;
        }
        const splitindex = msg.indexOf(':');
        if (splitindex == -1) {
            this.msgPlace.add(msg);
            return;
        }
        this.msgPlace.add(msg.substring(splitindex + 1), msg.substring(0, splitindex));
    });
    ssm.addListener('content.measureMode', () => this.check(true), false);
    ssm.addListener('content.connections', () => this.check(true), false);
    this.retryBtn = createHtml('button', createHtml('div', fieldset, { style: 'text-align:center'}), { type: 'button' }, 'Nochmal üben');
    this.retryBtn.style = 'display:none';
    this.retryBtn.addEventListener('click', () => ssm.setValue('content.reset', true));
    ssm.addListener('lms.lessonStatus', status => this.retryBtn.style = (['passed', 'failed', 'completed'].indexOf(status) == -1) ? 'display:none' : '');
}

MeasureResults.prototype.procInputChange = function() {
    const data = this.items.map(item => item.getValue());
    const raw = JSON.stringify(data);
    const save = raw.substring(1, raw.length - 1);
    this.ssm.setValue('content.inputs', save);
}

MeasureResults.prototype.check = function(onlyIfAuto) {
    if (this._timeout !== undefined)
        return;
    this._timeout = setTimeout(() => {
        const item = this.items[this.ssm.getValue('content.levelsDone')];
        if (item)
            item.check(onlyIfAuto);
        this._timeout = undefined;
    }, 0);
}

MeasureResults.Item = function(
    parent,
    data
) {
    this.parent = parent;
    this.node = createHtml('li', parent.listnode);
    const content = data.input ? createHtml('label', this.node) : this.node;
    content.innerHTML = data.html || '';
    if (data.input) {
        this.input = createHtml('input', content, { "type": data.input, "autocomplete": "off" });
        this.input.addEventListener('change', () => parent.procInputChange());
    }
    if (data.unit)
        createHtml('span', content).innerHTML = data.unit;
    if (!data.autosubmit)
        this.submitBtn = createHtml('button', content, { "type": "submit" }, "Überprüfen");
    if (data.autosubmit)
        this.autosubmit = true;
    this.checkers = (data.check instanceof Array) ? data.check : [];
}

MeasureResults.Item.prototype.updateState = function(
    state       // -1 → Invisibel, 0 → Active, 1 → Done, 2 → Failed
) {
    this.node.className = ['later', '', 'done', 'failed'][state + 1];
    if (this.input)
        this.input.disabled = (state != 0);
}

MeasureResults.Item.prototype.getValue = function() {
    return this.input ? this.input.value : 0;
}

MeasureResults.Item.prototype.setValue = function(value) {
    if (!this.input || this.input.value == value)
        return;
    this.input.value = value;
}

MeasureResults.Item.prototype.check = function(onlyIfAuto) {
    if (onlyIfAuto && !this.autosubmit)
        return;
    const input = this.getInput();
    this.parent.values.importValue('input', input.value, input.precision);
    MeasureStateChecker.run(this.checkers, this.parent.ssm, this.parent.simulation, this.parent.values, input.unit);
}

MeasureResults.Item.prototype.getInput = function() {
    if (!this.input)
        return { value: 0, precision: 0, unit: '' };
    const input = this.input.value;
    const parts = input.match(/^([+-]?\d+([\.,]\d+)?)\s*([a-zA-Z].*)?$/);
    if (!parts)
        return { value: 0, precision: 0, unit: '' };
    parts[1] = parts[1].replace(',', '.');
    let number = Number.parseFloat(parts[1]);
    let precision = Number.parseFloat(parts[1].replace('-', '').replace(/\d/g, '0').replace(/0$/, '1'));
    const unit = parts[3] ? parts[3] : '1';
    switch ((unit.length > 1) && unit.charAt(0)) {
        case 'p':
            number /= 100000000000;
            precision /= 100000000000;
            break;
        case 'n':
            number /= 100000000;
            precision /= 100000000;
            break;
        case 'µ':
            number /= 1000000;
            precision /= 1000000;
            break;
        case 'm':
            number /= 1000;
            precision /= 1000;
            break;
        case 'c':
            number /= 100;
            precision /= 100;
            break;
        case 'd':
            number /= 10;
            precision /= 10;
            break;
        case 'k':
            number *= 1000;
            precision *= 1000;
            break;
        case 'M':
            number *= 1000000;
            precision *= 1000000;
            break;
        case 'G':
            number *= 1000000000;
            precision *= 1000000000;
            break;
        case 'T':
            number *= 1000000000000;
            precision *= 1000000000000;
            break;
    }
    return { value: number, precision, unit };
}
